@@ -1,4 +1,4 @@
-#    Copyright 2019 Division of Medical Image Computing, German Cancer Research Center (DKFZ), Heidelberg, Germany
+#    Copyright 2020 Division of Medical Image Computing, German Cancer Research Center (DKFZ), Heidelberg, Germany
 #
 #    Licensed under the Apache License, Version 2.0 (the "License");
 #    you may not use this file except in compliance with the License.
@@ -14,17 +14,17 @@
 
 from collections import OrderedDict
 from batchgenerators.augmentations.utils import resize_segmentation
+from nnunet.configuration import default_num_threads, RESAMPLING_SEPARATE_Z_ANISO_THRESHOLD
 from nnunet.preprocessing.cropping import get_case_identifier_from_npz, ImageCropper
 from skimage.transform import resize
 from scipy.ndimage.interpolation import map_coordinates
 import numpy as np
-from nnunet.experiment_planning.configuration import RESAMPLING_SEPARATE_Z_ANISOTROPY_THRESHOLD
 from batchgenerators.utilities.file_and_folder_operations import *
 from multiprocessing.pool import Pool
 
 
-def get_do_separate_z(spacing):
-    do_separate_z = (np.max(spacing) / np.min(spacing)) > RESAMPLING_SEPARATE_Z_ANISOTROPY_THRESHOLD
+def get_do_separate_z(spacing, anisotropy_threshold=RESAMPLING_SEPARATE_Z_ANISO_THRESHOLD):
+    do_separate_z = (np.max(spacing) / np.min(spacing)) > anisotropy_threshold
     return do_separate_z
 
 
@@ -34,7 +34,8 @@ def get_lowres_axis(new_spacing):
 
 
 def resample_patient(data, seg, original_spacing, target_spacing, order_data=3, order_seg=0, force_separate_z=False,
-                     cval_data=0, cval_seg=-1, order_z_data=0, order_z_seg=0):
+                     cval_data=0, cval_seg=-1, order_z_data=0, order_z_seg=0,
+                     separate_z_anisotropy_threshold=RESAMPLING_SEPARATE_Z_ANISO_THRESHOLD):
     """
     :param cval_seg:
     :param cval_data:
@@ -48,6 +49,9 @@ def resample_patient(data, seg, original_spacing, target_spacing, order_data=3,
     /never resample along z separately
     :param order_z_seg: only applies if do_separate_z is True
     :param order_z_data: only applies if do_separate_z is True
+    :param separate_z_anisotropy_threshold: if max_spacing > separate_z_anisotropy_threshold * min_spacing (per axis)
+    then resample along lowres axis with order_z_data/order_z_seg instead of order_data/order_seg
+
     :return:
     """
     assert not ((data is None) and (seg is None))
@@ -69,10 +73,10 @@ def resample_patient(data, seg, original_spacing, target_spacing, order_data=3,
         else:
             axis = None
     else:
-        if get_do_separate_z(original_spacing):
+        if get_do_separate_z(original_spacing, separate_z_anisotropy_threshold):
             do_separate_z = True
             axis = get_lowres_axis(original_spacing)
-        elif get_do_separate_z(target_spacing):
+        elif get_do_separate_z(target_spacing, separate_z_anisotropy_threshold):
             do_separate_z = True
             axis = get_lowres_axis(target_spacing)
         else:
@@ -118,7 +122,7 @@ def resample_data_or_seg(data, new_shape, is_seg, axis=None, order=3, do_separat
     new_shape = np.array(new_shape)
     if np.any(shape != new_shape):
         if do_separate_z:
-            print("separate z")
+            print("separate z, order in z is", order_z, "order inplane is", order)
             assert len(axis) == 1, "only one anisotropic axis supported"
             axis = axis[0]
             if axis == 0:
@@ -167,12 +171,13 @@ def resample_data_or_seg(data, new_shape, is_seg, axis=None, order=3, do_separat
                             reshaped_multihot = np.round(
                                 map_coordinates((reshaped_data == cl).astype(float), coord_map, order=order_z,
                                                 cval=cval, mode='nearest'))
-                            reshaped[reshaped_multihot >= 0.5] = cl
+                            reshaped[reshaped_multihot > 0.5] = cl
                         reshaped_final_data.append(reshaped[None])
                 else:
                     reshaped_final_data.append(reshaped_data[None])
             reshaped_final_data = np.vstack(reshaped_final_data)
         else:
+            print("no separate z, order", order)
             reshaped = []
             for c in range(data.shape[0]):
                 reshaped.append(resize_fn(data[c], new_shape, order, cval=cval, **kwargs)[None])
@@ -196,6 +201,8 @@ class GenericPreprocessor(object):
         self.normalization_scheme_per_modality = normalization_scheme_per_modality
         self.use_nonzero_mask = use_nonzero_mask
 
+        self.resample_separate_z_anisotropy_threshold = RESAMPLING_SEPARATE_Z_ANISO_THRESHOLD
+
     @staticmethod
     def load_cropped(cropped_output_dir, case_identifier):
         all_data = np.load(os.path.join(cropped_output_dir, "%s.npz" % case_identifier))['data']
@@ -225,8 +232,13 @@ class GenericPreprocessor(object):
             'spacing_transposed': original_spacing_transposed,
             'data.shape (data is transposed)': data.shape
         }
+
+        # remove nans
+        data[np.isnan(data)] = 0
+
         data, seg = resample_patient(data, seg, np.array(original_spacing_transposed), target_spacing, 3, 1,
-                                     force_separate_z=force_separate_z, order_z_data=0, order_z_seg=0)
+                                     force_separate_z=force_separate_z, order_z_data=0, order_z_seg=0,
+                                     separate_z_anisotropy_threshold=self.resample_separate_z_anisotropy_threshold)
         after = {
             'spacing': target_spacing,
             'data.shape (data is resampled)': data.shape
@@ -246,8 +258,6 @@ class GenericPreprocessor(object):
         assert len(self.use_nonzero_mask) == len(data), "self.use_nonzero_mask must have as many entries as data" \
                                                         " has modalities"
 
-        print("normalization...")
-
         for c in range(len(data)):
             scheme = self.normalization_scheme_per_modality[c]
             if scheme == "CT":
@@ -280,7 +290,6 @@ class GenericPreprocessor(object):
                     mask = np.ones(seg.shape[1:], dtype=bool)
                 data[c][mask] = (data[c][mask] - data[c][mask].mean()) / (data[c][mask].std() + 1e-8)
                 data[c][mask == 0] = 0
-        print("normalization done")
         return data, seg, properties
 
     def preprocess_test_case(self, data_files, target_spacing, seg_file=None, force_separate_z=None):
@@ -314,7 +323,7 @@ class GenericPreprocessor(object):
             pickle.dump(properties, f)
 
     def run(self, target_spacings, input_folder_with_cropped_npz, output_folder, data_identifier,
-            num_threads=8, force_separate_z=None):
+            num_threads=default_num_threads, force_separate_z=None):
         """
 
         :param target_spacings: list of lists [[1.25, 1.25, 5]]
@@ -350,13 +359,95 @@ class GenericPreprocessor(object):
             p.join()
 
 
+class Preprocessor3DDifferentResampling(GenericPreprocessor):
+    def resample_and_normalize(self, data, target_spacing, properties, seg=None, force_separate_z=None):
+        """
+        data and seg must already have been transposed by transpose_forward. properties are the un-transposed values
+        (spacing etc)
+        :param data:
+        :param target_spacing:
+        :param properties:
+        :param seg:
+        :param force_separate_z:
+        :return:
+        """
+
+        # target_spacing is already transposed, properties["original_spacing"] is not so we need to transpose it!
+        # data, seg are already transposed. Double check this using the properties
+        original_spacing_transposed = np.array(properties["original_spacing"])[self.transpose_forward]
+        before = {
+            'spacing': properties["original_spacing"],
+            'spacing_transposed': original_spacing_transposed,
+            'data.shape (data is transposed)': data.shape
+        }
+
+        # remove nans
+        data[np.isnan(data)] = 0
+
+        data, seg = resample_patient(data, seg, np.array(original_spacing_transposed), target_spacing, 3, 1,
+                                     force_separate_z=force_separate_z, order_z_data=3, order_z_seg=1,
+                                     separate_z_anisotropy_threshold=self.resample_separate_z_anisotropy_threshold)
+        after = {
+            'spacing': target_spacing,
+            'data.shape (data is resampled)': data.shape
+        }
+        print("before:", before, "\nafter: ", after, "\n")
+
+        if seg is not None:  # hippocampus 243 has one voxel with -2 as label. wtf?
+            seg[seg < -1] = 0
+
+        properties["size_after_resampling"] = data[0].shape
+        properties["spacing_after_resampling"] = target_spacing
+        use_nonzero_mask = self.use_nonzero_mask
+
+        assert len(self.normalization_scheme_per_modality) == len(data), "self.normalization_scheme_per_modality " \
+                                                                         "must have as many entries as data has " \
+                                                                         "modalities"
+        assert len(self.use_nonzero_mask) == len(data), "self.use_nonzero_mask must have as many entries as data" \
+                                                        " has modalities"
+
+        for c in range(len(data)):
+            scheme = self.normalization_scheme_per_modality[c]
+            if scheme == "CT":
+                # clip to lb and ub from train data foreground and use foreground mn and sd from training data
+                assert self.intensityproperties is not None, "ERROR: if there is a CT then we need intensity properties"
+                mean_intensity = self.intensityproperties[c]['mean']
+                std_intensity = self.intensityproperties[c]['sd']
+                lower_bound = self.intensityproperties[c]['percentile_00_5']
+                upper_bound = self.intensityproperties[c]['percentile_99_5']
+                data[c] = np.clip(data[c], lower_bound, upper_bound)
+                data[c] = (data[c] - mean_intensity) / std_intensity
+                if use_nonzero_mask[c]:
+                    data[c][seg[-1] < 0] = 0
+            elif scheme == "CT2":
+                # clip to lb and ub from train data foreground, use mn and sd form each case for normalization
+                assert self.intensityproperties is not None, "ERROR: if there is a CT then we need intensity properties"
+                lower_bound = self.intensityproperties[c]['percentile_00_5']
+                upper_bound = self.intensityproperties[c]['percentile_99_5']
+                mask = (data[c] > lower_bound) & (data[c] < upper_bound)
+                data[c] = np.clip(data[c], lower_bound, upper_bound)
+                mn = data[c][mask].mean()
+                sd = data[c][mask].std()
+                data[c] = (data[c] - mn) / sd
+                if use_nonzero_mask[c]:
+                    data[c][seg[-1] < 0] = 0
+            else:
+                if use_nonzero_mask[c]:
+                    mask = seg[-1] >= 0
+                else:
+                    mask = np.ones(seg.shape[1:], dtype=bool)
+                data[c][mask] = (data[c][mask] - data[c][mask].mean()) / (data[c][mask].std() + 1e-8)
+                data[c][mask == 0] = 0
+        return data, seg, properties
+
+
 class PreprocessorFor2D(GenericPreprocessor):
     def __init__(self, normalization_scheme_per_modality, use_nonzero_mask, transpose_forward: (tuple, list), intensityproperties=None):
         super(PreprocessorFor2D, self).__init__(normalization_scheme_per_modality, use_nonzero_mask,
                                                 transpose_forward, intensityproperties)
 
     def run(self, target_spacings, input_folder_with_cropped_npz, output_folder, data_identifier,
-            num_threads=8, force_separate_z=None):
+            num_threads=default_num_threads, force_separate_z=None):
         print("Initializing to run preprocessing")
         print("npz folder:", input_folder_with_cropped_npz)
         print("output_folder:", output_folder)
@@ -387,7 +478,8 @@ class PreprocessorFor2D(GenericPreprocessor):
         }
         target_spacing[0] = original_spacing_transposed[0]
         data, seg = resample_patient(data, seg, np.array(original_spacing_transposed), target_spacing, 3, 1,
-                                     force_separate_z=force_separate_z, order_z_data=0, order_z_seg=0)
+                                     force_separate_z=force_separate_z, order_z_data=0, order_z_seg=0,
+                                     separate_z_anisotropy_threshold=self.resample_separate_z_anisotropy_threshold)
         after = {
             'spacing': target_spacing,
             'data.shape (data is resampled)': data.shape
